from pluggy import HookimplMarker
from pydantic import BaseModel, Field
from datetime import datetime, timedelta
import boto3
from loguru import logger
import concurrent.futures
from typing import Annotated
import threading


from core.plugins import Result


# Define a hookimpl (implementation of the contract)
hookimpl = HookimplMarker("opsbox")


class RDSProvider:
    """Plugin for gathering data related to AWS RDS instances.

    Attributes:
        S3 (boto3.client): The boto3 client for S3.
        credentials (dict): A dictionary containing AWS access key, secret access key, and region.
    """

    @hookimpl
    def grab_config(self):
        """Return the plugin's configuration."""

        class RDSConfig(BaseModel):
            """Configuration for the AWS RDS plugin."""

            aws_access_key_id: Annotated[str,Field(..., description="AWS access key ID", required=True)]
            aws_secret_access_key: Annotated[str,Field(..., description="AWS secret access key", required=True)]
            aws_region: Annotated[
                str | None, Field(description="AWS-Region", required=False, default=None)
            ]

        return RDSConfig

    @hookimpl
    def activate(self) -> None:
        """Initialize the plugin."""
        logger.trace("Activating the RDS plugin")

    @hookimpl
    def set_data(self, model: type[BaseModel]) -> None:
        """Set the data for the plugin based on the model."""
        logger.trace("Setting data for RDS plugin...")
        self.credentials = model.model_dump()

    @hookimpl
    def gather_data(self):
        """
        Gathers data related to AWS RDS instances.

        Returns:
            dict: A dictionary containing the gathered data in the following format:
                "input": {
                        "rds_instances": [
                            {
                                "InstanceIdentifier": "instance-id",
                                "InstanceType": "db.t2.micro",
                                "Region": "us-west-1a",
                                "AllocatedStorage": 20,
                                "CPUUtilization": 5.5,
                                "Connections": 150,
                                "StorageUtilization": 80
                            },
                            ...
                        ]
                    }
        """
        credentials = self.credentials

        logger.info(credentials["aws_region"])

        if credentials["aws_region"] is None:
            region_client = boto3.client(
                "ec2",
                aws_access_key_id=credentials["aws_access_key_id"],
                aws_secret_access_key=credentials["aws_secret_access_key"],
                region_name="us-west-1",
            )

            regions = [region["RegionName"] for region in region_client.describe_regions()["Regions"]]
            logger.info(f"Regions: {regions}")

        else:
            regions = credentials["aws_region"].split(",")

        region_threads = []  # List to store threads
        instance_data = []  # List to store instances
        snapshots = []  # List to store snapshots


        def get_rds_instances(rds_client: boto3.client):
            """Retrieve information about RDS instances."""
            instances = []
            paginator = rds_client.get_paginator("describe_db_instances")
            for page in paginator.paginate():
                for instance in page["DBInstances"]:
                    instances.append(
                        {
                            "InstanceIdentifier": instance["DBInstanceIdentifier"],
                            "InstanceType": instance["DBInstanceClass"],
                            "Region": instance["AvailabilityZone"],
                            "AllocatedStorage": instance["AllocatedStorage"],
                            "Engine": instance["Engine"],
                        }
                    )
            return instances

        def get_rds_snapshots(rds_client: boto3.client):
            """Retrieve information about RDS snapshots."""
            snapshots = []
            paginator = rds_client.get_paginator("describe_db_snapshots")
            for page in paginator.paginate():
                for snapshot in page["DBSnapshots"]:
                    snapshots.append(
                        {
                            "SnapshotIdentifier": snapshot["DBSnapshotIdentifier"],
                            "InstanceIdentifier": snapshot["DBInstanceIdentifier"],
                            "SnapshotCreateTime": snapshot["SnapshotCreateTime"].isoformat(),
                            "AllocatedStorage": snapshot["AllocatedStorage"],
                            "StorageType": snapshot["StorageType"],
                        }
                    )
            return snapshots

        def get_cloudwatch_metric(
            instance_id: str, metric_name: str, statistic: str, period: int, cloudwatch_client: boto3.client
        ):
            """Retrieve CloudWatch metrics for a given instance."""
            end_time = datetime.now()
            start_time = end_time - timedelta(days=7)

            response = cloudwatch_client.get_metric_statistics(
                Namespace="AWS/RDS",
                MetricName=metric_name,
                Dimensions=[{"Name": "DBInstanceIdentifier", "Value": instance_id}],
                StartTime=start_time,
                EndTime=end_time,
                Period=period,
                Statistics=[statistic],
            )

            datapoints = response["Datapoints"]
            return sum(datapoint[statistic] for datapoint in datapoints) / len(datapoints) if datapoints else 0

        def get_storage_utilization(instance_id, cloudwatch_client, allocated_storage):
            """Calculate the storage utilization for a given instance."""
            avg_free_storage = get_cloudwatch_metric(
                instance_id, "FreeStorageSpace", "Average", 86400, cloudwatch_client
            )
            avg_free_storage_gb = avg_free_storage / (1024**3)  # Convert bytes to GB
            used_storage_gb = allocated_storage - avg_free_storage_gb
            return round((used_storage_gb / allocated_storage) * 100)

        # Initialize boto3 clients for RDS and CloudWatch




        def process_region(region):
            rds_client = boto3.client(
                "rds",
                aws_access_key_id=self.credentials["aws_access_key_id"],
                aws_secret_access_key=self.credentials["aws_secret_access_key"],
                region_name=region,
            )

            cloudwatch_client = boto3.client(
                "cloudwatch",
                aws_access_key_id=self.credentials["aws_access_key_id"],
                aws_secret_access_key=self.credentials["aws_secret_access_key"],
                region_name=region,
            )
            instances = get_rds_instances(rds_client)
            snapshots.append(get_rds_snapshots(rds_client))
        # Retrieve RDS instances

            

            # Use a ThreadPoolExecutor to gather metrics concurrently
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = []
                for instance in instances:
                    futures.append(
                        executor.submit(
                            get_cloudwatch_metric,
                            instance["InstanceIdentifier"],
                            "CPUUtilization",
                            "Average",
                            3600,
                            cloudwatch_client,
                        )
                    )
                    futures.append(
                        executor.submit(
                            get_cloudwatch_metric,
                            instance["InstanceIdentifier"],
                            "DatabaseConnections",
                            "Sum",
                            86400,
                            cloudwatch_client,
                        )
                    )
                    futures.append(
                        executor.submit(
                            get_storage_utilization,
                            instance["InstanceIdentifier"],
                            cloudwatch_client,
                            instance["AllocatedStorage"],
                        )
                    )

                # Collect results and populate instance data
                for i, instance in enumerate(instances):
                    instance["CPUUtilization"] = futures[i * 3].result()
                    instance["Connections"] = futures[i * 3 + 1].result()
                    instance["StorageUtilization"] = futures[i * 3 + 2].result()
                    instance_data.append(instance)


        for region in regions:
            region_thread = threading.Thread(target=process_region, args=(region,))
            region_threads.append(region_thread)
            region_thread.start()

        # Wait for all threads to complete
        for region_thread in region_threads:
            region_thread.join()


        rego_ready_data = {
            "input": {
                "rds_instances": instance_data,
                "rds_snapshots": snapshots,
                "date": datetime.now().isoformat() + "Z",
            }
        }
        item = Result(
            relates_to="rds",
            result_name="rds_instances",
            result_description="RDS Instances",
            details=rego_ready_data,
            formatted="",
        )
        logger.trace(f"RDS data: {item.details}")
        logger.success("Successfully gathered data for RDS instances")
        return item



from moto import mock_aws
import boto3
from .rds_provider import RDSProvider
from core.plugins import Result
from pydantic import BaseModel
from loguru import logger

# ruff: noqa: S101
@mock_aws
def test_rds_provider_gather_data(json_output=False):
    """Test of the RDS provider gather_data method.
    
    Args:
        json_output (bool, optional): If True, the test will output the JSON result to a file. Defaults to False.
            File will be saved to tests/data/rds_test_data.json.
    """
    
    # Mock AWS credentials (moto uses dummy credentials)
    aws_access_key_id = "fake_access_key"
    aws_secret_access_key = "fake_secret_key"  # noqa: S105
    aws_region = "us-west-1"

    # Create mock RDS client
    rds_client = boto3.client("rds", region_name=aws_region)

    # Create mock RDS instance
    rds_client.create_db_instance(
        DBInstanceIdentifier="test-instance",
        AllocatedStorage=20,
        DBInstanceClass="db.t2.micro",
        Engine="mysql",
        MasterUsername="admin",
        MasterUserPassword="password",
        AvailabilityZone=aws_region + "a",
    )

    # Create mock RDS snapshot
    rds_client.create_db_snapshot(
        DBSnapshotIdentifier="test-snapshot",
        DBInstanceIdentifier="test-instance",
    )

    # Instantiate provider and set credentials
    provider = RDSProvider()

    class MockCredentials(BaseModel):
        aws_access_key_id: str
        aws_secret_access_key: str
        aws_region: str | None

    credentials = MockCredentials(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_region=aws_region,
    )

    provider.set_data(credentials)

    # Disable logging for test clarity
    logger.remove()
    logger.add(lambda msg: None)

    # Invoke gather_data method
    result = provider.gather_data()
    if json_output:
        import json

        with open("tests\data\\rds_test_data.json", "w") as f:
            json.dump(result.model_dump()["details"]["input"], f, indent=4)

    # Assertions
    assert isinstance(result, Result)
    assert result.relates_to == "rds"
    assert result.result_name == "rds_instances"
    details = result.details["input"]
    assert "rds_instances" in details
    assert "rds_snapshots" in details

    # Verify RDS instances data
    instances = details["rds_instances"]
    assert len(instances) == 1
    instance = instances[0]
    assert instance["InstanceIdentifier"] == "test-instance"
    assert instance["InstanceType"] == "db.t2.micro"
    assert instance["Engine"] == "mysql"
    assert instance["AllocatedStorage"] == 20
    # Metrics may not be available in moto's mock, so check for default values
    assert "CPUUtilization" in instance
    assert "Connections" in instance
    assert "StorageUtilization" in instance

    # Verify RDS snapshots data
    snapshots = details["rds_snapshots"]
    assert len(snapshots) == 1
    snapshot_list = snapshots[0]
    assert len(snapshot_list) == 1
    snapshot = snapshot_list[0]
    assert snapshot["SnapshotIdentifier"] == "test-snapshot"
    assert snapshot["InstanceIdentifier"] == "test-instance"
